/* 
 * Copyright (C) 2003 Alex Zolotov
 * Mucked with by Tugrul Galatali <tugrul@galatali.com>
 *
 * MatrixView is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as 
 * published by the Free Software Foundation.
 *
 * MatrixView is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef MATRIXVIEW_TEXTURES_H
#define MATRIXVIEW_TEXTURES_H

extern unsigned char *cfont;
extern unsigned int cfont_size;
extern unsigned int cfont_compressedsize;
extern unsigned char *cpics;
extern unsigned int cpics_size;
extern unsigned int cpics_compressedsize;

#endif
