#ifndef SMOKETEX_H
#define SMOKETEX_H

#define SMOKETEXSIZE 64

extern unsigned char *presmoke1;
extern unsigned int presmoke1_size;
extern unsigned int presmoke1_compressedsize;
extern unsigned char *presmoke2;
extern unsigned int presmoke2_size;
extern unsigned int presmoke2_compressedsize;
extern unsigned char *presmoke3;
extern unsigned int presmoke3_size;
extern unsigned int presmoke3_compressedsize;
extern unsigned char *presmoke4;
extern unsigned int presmoke4_size;
extern unsigned int presmoke4_compressedsize;
extern unsigned char *presmoke5;
extern unsigned int presmoke5_size;
extern unsigned int presmoke5_compressedsize;

#endif // SMOKETEX_H
