#ifndef MOONTEX_H
#define MOONTEX_H

#define MOONTEXSIZE 64

extern unsigned char *moontint;
extern unsigned int moontint_size;
extern unsigned int moontint_compressedsize;
extern unsigned char *moonalpha;
extern unsigned int moonalpha_size;
extern unsigned int moonalpha_compressedsize;

#endif
