/*
 * Copyright (C) 2002  Terence M. Welsh
 * Ported to Linux by Tugrul Galatali <tugrul@galatali.com>
 *
 * Lattice is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as 
 * published by the Free Software Foundation.
 *
 * Lattice is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef LATTICE_TEXTURE_H
#define LATTICE_TEXTURE_H

extern unsigned char *indtex1;
extern unsigned int indtex1_size;
extern unsigned int indtex1_compressedsize;
extern unsigned char *indtex2;
extern unsigned int indtex2_size;
extern unsigned int indtex2_compressedsize;
extern unsigned char *crystex;
extern unsigned int crystex_size;
extern unsigned int crystex_compressedsize;
extern unsigned char *chrometex;
extern unsigned int chrometex_size;
extern unsigned int chrometex_compressedsize;
extern unsigned char *brasstex;
extern unsigned int brasstex_size;
extern unsigned int brasstex_compressedsize;
extern unsigned char *shinytex;
extern unsigned int shinytex_size;
extern unsigned int shinytex_compressedsize;
extern unsigned char *doughtex;
extern unsigned int doughtex_size;
extern unsigned int doughtex_compressedsize;
extern unsigned char *ghostlytex;
extern unsigned int ghostlytex_size;
extern unsigned int ghostlytex_compressedsize;
extern unsigned char *circuittex;
extern unsigned int circuittex_size;
extern unsigned int circuittex_compressedsize;

#endif
