#ifndef EARTHTEX_H
#define EARTHTEX_H

#define EARTHNEARSIZE 256
#define EARTHFARSIZE 512

extern unsigned char *earthnearmap;
extern unsigned int earthnearmap_size;
extern unsigned int earthnearmap_compressedsize;
extern unsigned char *earthfarmap;
extern unsigned int earthfarmap_size;
extern unsigned int earthfarmap_compressedsize;
extern unsigned char *earthlightmap;
extern unsigned int earthlightmap_size;
extern unsigned int earthlightmap_compressedsize;

#endif
