/*
 * Copyright (C) 2002 Andreas Gustafsson
 * Ported to Linux by Tugrul Galatali <tugrul@galatali.com>
 *
 * Euphoria is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as 
 * published by the Free Software Foundation.
 *
 * Euphoria is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef HUFO_TUNNEL_TEXTURES_H
#define HUFO_TUNNEL_TEXTURES_H

extern unsigned char *swirlmap;
extern unsigned int swirlmap_size;
extern unsigned int swirlmap_compressedsize;
extern unsigned char *marblemap;
extern unsigned int marblemap_size;
extern unsigned int marblemap_compressedsize;

#endif
